/*
*
*	    Written by Gaute Gamnes, NTNU
*       spring of 2014. Date: 140521
*
*       /* Code borrowed and modified from several sources...
* 		Serial information:
* 		http://en.wikibooks.org/wiki/Serial_Programming/termios
* 		http://blog.eduardofleury.com/archives/2007/11/16
* 		MAVLink: as well as serial:
* 
http://diydrones.com/forum/topics/need-help-for-serial-comunication?id=705844%3ATopic%3A1038239&page=1#comments
*
*       Compile with:
*       gcc main.c -o programExe -I ../../mavlink/Hello-World/mavlink/include/mavlink/v1.0/common
*
*       Before running the program UART needs to be enabled on the BeagleBone.
*       UART 4 is used in this application; Pin 11(Rx) and Pin 13(Tx) on Header P9.
*       To enable UART 4: run command
*       echo BB-UART4 > /sys/devices/bone_capemgr.*BACKSLASHslots
*       The BACKSLASH is a / but can not be inserted in the comment section of this code.
*
*		Using MAVLink: Download the newest .zip file from https://github.com/mavlink/mavlink/downloads
*		Unzip on the BeagleBone Black and include <mavlink.h> in the code.
*		When compiling the program, include the -I flag during compile and append the path to the
*		...../include/mavlink/v1.0/common folder.
*		If the MAVLink headers are included in the PATH environment of the BeagleBone it might not be needed with
*		the compile flag.
*
*/

#include <termios.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <sys/ioctl.h>

#include <mavlink.h>

#define TTY_SERIAL_PORT "/dev/ttyO4" // Path to serial port, UART 4, pin 11(Rx) and 13(Tx)

int init_serial_communication(char* port) {
	int tty_fd = 0; // The returned file handle for the device
	tty_fd = open(port, O_RDWR | O_NOCTTY | O_NDELAY);
	// O_RDWR - opens port for read and write
	// O_NOCTTY - The port never becomes the controlling terminal of the process
	// O_NDELAY - Non-blocking IO
	if(tty_fd == -1) {
		printf("Failed to open port\n");
		return tty_fd;
	}

	struct termios serial_options, original_serial_options;
	// Get current terminal settings
	if(tcgetattr(tty_fd, &original_serial_options) == -1) {
		printf("Error getting tty attributes %s - %s(%d).\n", tty_fd, strerror(errno), errno);
		return -1;
	}
	serial_options = original_serial_options; // Copy settings to new and play with them

	// Set termianl to something like raw mode
	//Input is availa char by char, echoing is disabled, no special processing of IO char
	cfmakeraw(&serial_options);
	// Set input and output speed?
	cfsetspeed(&serial_options, 57600);
	// 8N1
	serial_options.c_cflag &= ~PARENB;
	serial_options.c_cflag &= ~CSTOPB;
	serial_options.c_cflag &= ~CSIZE;
	serial_options.c_cflag |= CS8;
	// no flow control
	serial_options.c_cflag &= ~CRTSCTS;
	serial_options.c_cflag |= CREAD | CLOCAL;  // turn on READ & ignore ctrl lines
	serial_options.c_iflag &= ~(IXON | IXOFF | IXANY); // turn off s/w flow ctrl
	serial_options.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG); // make raw
	serial_options.c_oflag &= ~OPOST; // make raw
	// see: http://unixwiz.net/techtips/termios-vmin-vtime.html
	serial_options.c_cc[VMIN]  = 0;
	serial_options.c_cc[VTIME] = 20;

	// Apply settings
	if(tcsetattr(tty_fd, TCSANOW, &serial_options) == -1) {
		printf("Error setting tty attributes %s - %s(%d).\n", tty_fd, strerror(errno), errno);
		return -1;
	}
	// clear all flags on descriptor, enable direct IO
	if(fcntl(tty_fd, F_SETFL, 0)) {
			printf("Error setting tty non-blocking %s - %s(%d).\n", tty_fd, strerror(errno), errno);
			exit(EXIT_FAILURE);
	}

	return tty_fd;
}

int main(int argc, char* argv) {
	const char *dev = "/dev/ttyO4"; // Path to serial port, UART 4, pin 11 and 13

	int tty_fd = init_serial_communication(TTY_SERIAL_PORT);
	if(tty_fd == -1) {
		printf("Error opening serial port!\n");
		exit(EXIT_FAILURE);
	}

	/* Receive information from serial port */
	char rec_byte;
	//NEEDS MAVLINK for this to work
	mavlink_status_t msg_status;
	mavlink_message_t msg;
	static const int FIXED_HEADER_LEN = MAVLINK_NUM_HEADER_BYTES + MAVLINK_NUM_CHECKSUM_BYTES;
	int serial_fd = tty_fd; //would have to be used if sent into a thread, thus taking the value of args.
	while(1) {
		int rdChar = read(serial_fd, &rec_byte, 1);
		if(rdChar == -1) {
				perror("Error reading serial port");
		}
		if(rdChar == 0) {
				usleep(10*1000); //wait 10msec try again
				continue;
		}
		// Read the packet that is receive using the MAVLink provided function parse char.
		if ( mavlink_parse_char(MAVLINK_COMM_0, rec_byte, &msg, &msg_status) ) {
			//Packet received
			printf("Serial -> BBB: SYS: %d, COMP: %d, LEN: %d, MSG ID: %d\n", 
					msg.sysid, msg.compid, msg.len, msg.msgid);
			// Print the packet
			int j;
			for (j=0; j<msg.len + FIXED_HEADER_LEN; j++) {
				printf("0x%02x ", *((uint8_t*)&msg.magic + j) & 0x00ff);
				if(((j+1) % 8) == 0) printf("\n");
			}
			// To tidy up the print outs
			if(! ((j+1) % 8) == 0) printf("\n");

			//Switch case to handle messages,
			// Take mavlink_message_t* msg as input.
			switch(msg.msgid) {
				case MAVLINK_MSG_ID_HEARTBEAT:
					printf("Heartbeat received\n");
					break;
				case MAVLINK_MSG_ID_COMMAND_LONG:
					printf("Command long received\n");
					break;
				case MAVLINK_MSG_ID_GPS_RAW_INT:
					printf("GPS raw int received\n");
					break;
				case MAVLINK_MSG_ID_GPS_STATUS:
					printf("GPS status received\n");
					break;
				case MAVLINK_MSG_ID_RAW_IMU: //#27
					printf("Raw IMU received\n");
					break;
				case MAVLINK_MSG_ID_SCALED_PRESSURE: //#29
					printf("Scaled pressure received\n");
					break;
				case MAVLINK_MSG_ID_ATTITUDE: //#30
					printf("Attitude received\n");
					mavlink_attitude_t packet;
					mavlink_msg_attitude_decode(&msg, &packet);
					printf("Pitch: %f, yaw: %f, roll: %f \n", 
						packet.pitch, packet.yaw, packet.roll);
					break;
				case MAVLINK_MSG_ID_GLOBAL_POSITION_INT: //#33, filtered GPS
					printf("Global position int received\n");
					mavlink_global_position_int_t packet2;
					mavlink_msg_global_position_int_decode(&msg, &packet2);
					printf("Latitude: %d, Longitude: %d, Altitude: %d \n",
							packet2.lat, packet2.lon, packet2.alt);
					//Lat and lon expressed as *1E7. Alt in meters, * 1000 (mm).
					break;
				case MAVLINK_MSG_ID_VFR_HUD: //#74, Metrics for fixed wing
					printf("VFR_HUD received\n");
					break;
				default:
					//Do nothing
					printf("MSG ID not handled\n");
			}
		}
	}

	/* Done receiving from serial port */

	// Restore terminal settings NO NEED in this particular case, cause program never ends
	//tcsetattr(0, TCSANOW, &original_serial_config);

	close(tty_fd);

	return 0;
}





