/*
 * Find face in webcam frame with or without Blob detection
 *
 * main.cpp
 * dependencies: funcTimer.h and .cpp for run-time calculations
 * 	; and the cascade files for the OpenCV function needs to be in the same folder as the run file
 *
 *  Created on: 15mai2014
 *      Author: Gaute Gamnes
 */


/*
 * Some code is adapted from http://stackoverflow.com/questions/8166024/how-does-findcontours-cycle-through-the-image-opencv-2-3
 * Other code is adapted from Vegard Hammerseth, 2013.
 */
#include <stdio.h>
#include <iostream>
#include <fstream>
#include <string>
#include <time.h>
#include <sstream>
#include <unistd.h>

extern "C" {
	#include <pthread.h>
}

#include <opencv2/core/core.hpp>
#include <opencv2/features2d/features2d.hpp>
#include <opencv2/nonfree/nonfree.hpp>
#include <opencv2/calib3d/calib3d.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/objdetect/objdetect.hpp>
#include <opencv2/contrib/contrib.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/stitching/stitcher.hpp>
using namespace std;
using namespace cv;

#include "funcTimer.h"

//#define GUI or BBB mode
#define GRAPHICS
#define DEBUG


#ifdef DEBUG
#define DB(x) x
#else
#define DB(x)
#endif

// Extracts the red channel of the image and removes some of the red from other colors
void extractRed(const cv::Mat& src, cv::Mat& dst) {
   //Get red channel from src
   cv::Mat colorChannel(src.rows, src.cols, CV_8UC1);
   int fromTo[] = {2,0};
   cv::mixChannels(&src, 1, &colorChannel, 1, fromTo, 1);

   //Get grayscale of src
   cv::Mat gray(src.rows, src.cols, CV_8UC1);
   cv::cvtColor(src, gray, CV_BGR2GRAY);

   //Remove other colors than "real" red (white is also removed for instance)
   cv::Mat pureColor(src.rows, src.cols, CV_8UC1);
   pureColor = colorChannel - gray;

   dst = pureColor.clone();
}

// Finds the largest Contour in a vector of contours and saves the index in maxIndex.
// Returns the area of the largest contour.
double findLargestContour(const vector< vector<cv::Point> > &contours, int *maxIndex) {
   double area;
   double maxArea = 0;

   for(size_t i=0; i<contours.size(); i++) {
       area = contourArea(contours[i]);
       //cout << "Area of contour: " << area << endl; //For debugging
       if(area > maxArea) {
           maxArea = area;
           *maxIndex = i;
       }
   }

   return maxArea;
}


// Find a face in an image. Can also be used for finding other features by choosing
// which cascade file to load. The result is shown with a green circle.
int findFaceInImage(Mat &img, Mat &result) {
	vector<Rect> found, found_filtered;

	CascadeClassifier face_cascade;
	string cascadeFileName1 = "hogcascade_pedestrians.xml";
	string cascadeFileName2 = "haarcascade_frontalface_alt.xml";
	string cascadeFileName3 = "haarcascade_fullbody.xml";
	string cascadeFileName4 = "haarcascade_lowerbody.xml";
	string cascadeFileName5 = "haarcascade_mcs_upperbody.xml";
	string cascadeFileName6 = "haarcascade_upperbody.xml";

	if( !face_cascade.load(cascadeFileName2) ) {
		cout << "Failed loading cascade file" << endl;
		exit(0);
	}

	//Parameters for cascade detection
	double scaleFactor = 1.1; //How much the image size is reduced each image scale
	int minNeighbors = 2; //How many neighbor each candidate rectangle should have to retain it
	int flags = 0|CV_HAAR_SCALE_IMAGE; //Not used in new cascade
	Size minSize = Size(8,8); //Minimum possible object size
	Size maxSize = Size(32,32); //Maximum possible object size

	Size s = img.size();
	cout << "Detecting faces... on img of size: " << s.height << " " << s.width << endl;
	face_cascade.detectMultiScale(img, found, scaleFactor, minNeighbors, flags, minSize);//, maxSize);
	cout << "Number of objects found: " << found.size() << endl;

	for(size_t j=0; j<found.size(); j++) {
		Point center( found[j].x + found[j].width*0.5, found[j].y + found[j].height*0.5 );
		cout << "Object found at; x: " << found[j].x + found[j].width*0.5 << " and y: " << found[j].y + found[j].height*0.5 << endl;
		int radius = cvRound( (found[j].width + found[j].height)*0.25 );
		circle(result, center, radius, Scalar(0,255,0), 4, 8, 0);
	}

	return found.size();
}

// Data used for communication between the threads
struct thread_data_t {
	Mat frame;
	int test;
	int algorithmModeSelector;
}GLOB_thread_data_t;

pthread_mutex_t frameLock;
pthread_barrier_t getStartedBarrier;

// Retrieval thread. Gets the frames from the webcamera
void * getWebcamFrames(void * argument) {
	//thread_data_t *data = (reinterpret_cast<thread_data_t*>(argument));
	VideoCapture capture(0); // 0 for standard camera, 1 for other. If laptop has camera, might need to use 1.
	if(!capture.isOpened()) {
		cout << "Cannot open the video capture" << endl;
		exit(0);
	}

	capture.set(CV_CAP_PROP_FRAME_HEIGHT, 720); //set camera rate
	capture.set(CV_CAP_PROP_FRAME_WIDTH, 1280); //set camera rate
	capture >> GLOB_thread_data_t.frame;
	pthread_barrier_wait(&getStartedBarrier);

	while(1) {
		pthread_mutex_lock(&frameLock);
		capture >> GLOB_thread_data_t.frame;
		pthread_mutex_unlock(&frameLock);
		usleep(100);
		//pthread_yield(); //give up CPU
		//WITHOUT SLEEP OTHER THREAD HAS TO WAIT FOR A LONG TIME BEFORE GAINING ACCESS TO IMAGE!
	}
	pthread_exit(NULL);
	return 0;
}

// Threshold values for The binary segmentation. Might need to be altered to fit the red value that is desired
#define THRESHOLD_RED 50
#define MAX_BINARY_VALUE 255

// The algorithm which find a face in the webcamera frame. It can choose to perform the face detection on the entire image,
// or, do the face detection on the segmented image. Chosen by Argument.
void * objectAlgorithm(void * argument) {
	//thread_data_t *data = (reinterpret_cast<thread_data_t*>(argument));
	pthread_barrier_wait(&getStartedBarrier);
	DB( funcTimer duration; )

	DB( int saveImageCounter = 0; )
	DB( string fileName;
		ofstream myFile; )


	int modeSelector_FindRedBlobs = GLOB_thread_data_t.algorithmModeSelector;
	//USE this for selection if blob detection is to be performed or not. 1 => use blob, 0 => no blob.

	if(!modeSelector_FindRedBlobs) myFile.open("/media/mmc0/noBlob/timeResults.txt");
	else myFile.open("/media/mmc0/blob/timeResults.txt");


	while(1) {
		//Get frame from webcam
		DB( duration.timerStart(); ) // Timer start
		pthread_mutex_lock(&frameLock);
		Mat frame = GLOB_thread_data_t.frame.clone();
		pthread_mutex_unlock(&frameLock);
		DB( duration.timerEnd(); )   // Timer end
		DB( cout << "!!!Get frame wait: " << duration.getTimeInMS() << endl; )

		// Keep track of the images:
		DB( stringstream ss;
			ss << saveImageCounter;
			saveImageCounter++; )

		// Process image: SELECT BETWEEN BLOB OR NOT BLOB DETECTION
		if(!modeSelector_FindRedBlobs) {

			Mat faceResult;
			faceResult = frame.clone();

			DB( duration.timerStart(); ) // Timer start
			if(findFaceInImage(frame, faceResult)) {
				DB( duration.timerEnd(); )   // Timer end
				DB( cout << duration.getTimeInMS() << endl; )

				DB( cout << "Face found" << endl; )

				DB(		fileName = "/media/mmc0/noBlob/result_" + ss.str() + ".jpg";
						imwrite(fileName, faceResult);
						myFile << "Time used on file: " << fileName << " \t was: " << duration.getTimeInMS() << endl; )
			}
		} else if (modeSelector_FindRedBlobs) {
			DB( duration.timerStart(); ) // Timer start

			//Save Frame
			DB( fileName = "/media/mmc0/blob/Frame_" + ss.str() + ".jpg";
				imwrite(fileName, frame); )

			//Extract red blobs
			DB( cout << "### Extract red blobs" << endl;)
			Mat redMat;
			extractRed(frame, redMat);

			//Threshold into binary
			DB( cout << "### Threshold into binary" << endl; )
			cv::threshold(redMat, redMat, THRESHOLD_RED, MAX_BINARY_VALUE, THRESH_BINARY);

			//Save binary image
			DB(		fileName = "/media/mmc0/blob/Binary_" + ss.str() + ".jpg";
					imwrite(fileName, redMat); )

			//Extract contours
			DB(cout << "### Extract contours" << endl;)
			vector< vector<cv::Point> > contours;
			vector<Vec4i> hierarchy;
			findContours(redMat, contours, hierarchy, CV_RETR_TREE, CV_CHAIN_APPROX_SIMPLE, Point(0,0) );
			DB( cout << "Contours found: " << contours.size() << endl; )

			//Find largest contour
			DB(cout << "### Find largest contour" << endl;)
			int largestContourArea = -1;
			int largestContourIndex = -1;
			if(contours.size() != 0) {
				largestContourArea = findLargestContour(contours, &largestContourIndex);
				DB( cout << "Largest contour index: " << largestContourIndex << endl; )
			}

			// Process contour area:
			DB(cout << "### Process contour area" << endl;)
			if(largestContourIndex != -1) {
				DB( cout << "Object found. Size: " << largestContourArea << endl; )
				// Extract contour area from frame
				cv::Mat imgROI;
				cv::Rect rect = cv::boundingRect(contours[largestContourIndex]);

				DB ( cout << "Extracting ROI from image, h: " << rect.height << " w: " << rect.width << endl; )
				//Since the object used was only wearing a red sweater,
				//the ROI needs to be expanded somewhat larger in order to do the detection of a face
				cv::Rect newRect;
				newRect = rect + cv::Size(rect.area()*0.05,rect.area()*0.05);
				Size imgSize = frame.size();
				// Make sure the ROI is contained to the original image
				if(newRect.height > imgSize.height) newRect.height = imgSize.height;
				if(newRect.width > imgSize.width) newRect.width = imgSize.width;

				// Set location of rect
				newRect.x -= newRect.width/2;
				newRect.y -= newRect.height/2;

				// Make sure the ROI stays within the boundaries of the original image
				if(newRect.x < 0) newRect.x = 0; // Test left side
				if(newRect.y < 0) newRect.y = 0; // Test top
				if( (newRect.x + newRect.width) > imgSize.width ) newRect.width -= ( (newRect.x + newRect.width) - imgSize.width);
				if( (newRect.y + newRect.height) > imgSize.height ) newRect.height -= ( (newRect.y + newRect.height) - imgSize.height);

				DB ( cout << "Extracting ROI from image, h: " << newRect.height << " w: " << newRect.width << endl; )
				imgROI = frame(newRect);
				Mat result = imgROI.clone();

				//Save ROI
				DB( 	fileName = "/media/mmc0/blob/ROI_" + ss.str() + ".jpg";
						imwrite(fileName, imgROI); )

				// Do the actual processing
				if(largestContourArea > 100) {
					DB( cout << "Object found and above threshold" << endl; )
					if(findFaceInImage(imgROI, result)) {
						DB( duration.timerEnd(); )   // Timer end
						DB( cout << duration.getTimeInMS() << endl; )

						DB( cout << "Face found, image saved" << endl; )
						DB( 	fileName = "/media/mmc0/blob/result_" + ss.str() + ".jpg";
								imwrite(fileName, result);
								myFile << "Time used on file: " << fileName << " \t was: " << duration.getTimeInMS() << endl; )
					}
				}
			}
		}
	}
	pthread_exit(NULL);
	return 0;
}

// Starts the two threads and never ends.
// Argument 1 is the mode selector between blob or no blob detection.
int main(int argc, char** argv) {
	if(argc < 2) {
		cout << "Arguments needs to be input to this program. " << endl;
		cout << "Arg 1: mode selector between blob (1) or no blob (0) detection" << endl;
		exit(0);
	}
	int modeSelector = atoi(argv[1]);
	if( (modeSelector == 0) || (modeSelector == 1)) {
		GLOB_thread_data_t.algorithmModeSelector = atoi(argv[1]);
	}
	else {
		cout << "Error in input argument 1. Needs to be 1 for blob, or 0 for no blob" << endl;
		exit(0);
	}

	pthread_t getImageFromWebcam, objectAlgorithm_t;

	pthread_mutex_init(&frameLock, NULL);
	pthread_barrier_init(&getStartedBarrier, NULL, 2);

	int createThread1 = pthread_create(&getImageFromWebcam, NULL, getWebcamFrames, NULL);//reinterpret_cast<void*>(&thread_data_t));
	if(createThread1 != 0) exit(EXIT_FAILURE);
	int createThread2 = pthread_create(&objectAlgorithm_t, NULL, objectAlgorithm, NULL);//reinterpret_cast<void*>(&thread_data_t));
	if(createThread2 != 0) exit(EXIT_FAILURE);

	pthread_join(getImageFromWebcam, NULL);
	pthread_join(objectAlgorithm_t, NULL);

	return 0;
}
